#include <stdio.h>

typedef struct
{
	char field1;
	char field2;
	long field3;
	short field4;
	char field5;
	int field6;
	char field7;
}
StructType;

/*
 * Poravnanje primitivnih tipova jednako je njihovoj velicini izrazenoj u broju bajtova.
 *
 * Poravnanje strukture jednako je poravnanju njenog polja sa najstrozim poravnanjem.
 * Velicina strukture (sizeof) mora biti celobrojni umnozak njenog poravnanja.
 *
 * Polje field3 tipa long structure StructType predstavlja njeno polje sa 
 * najstrozim poravnanjem (alignment) koje iznosi 8. Prema tome, poravnanje strukture
 * tipa StructType iznosi 8, dok je njena velicina celobrojni umnozak broja 8.
 *
 * Legenda predstave adresnog prostora:
 * 
 *                         +0       +1       +2       +3       +4       +5       +6       +7
 *
 * address + 0x08 ---> | 1 Byte | 1 Byte | 1 Byte | 1 Byte | 1 Byte | 1 Byte | 1 Byte | 1 Byte |
 * address + 0x00 ---> | 1 Byte | 1 Byte | 1 Byte | 1 Byte | 1 Byte | 1 Byte | 1 Byte | 1 Byte |
 *
 *
 *
 * Smestanje strukture tipa StructType u adresni prostor:
 *
 *                     |        |        |        |        |        |        |        |        |
 *                      -------- -------- -------- -------- -------- -------- -------- --------
 *                     | field7 |  padd  |  padd  |  padd  |  padd  |  padd  |  padd  |  padd  |
 *                     | field4 | field4 | field5 |  padd  | field6 | field6 | field6 | field6 |
 *                     | field3 | field3 | field3 | field3 | field3 | field3 | field3 | field3 |
 * struct_address ---> | field1 | field2 |  padd  |  padd  |  padd  |  padd  |  padd  |  padd  |
 *                      -------- -------- -------- -------- -------- -------- -------- --------
 *                     |        |        |        |        |        |        |        |        |
 */


// https://gcc.gnu.org/onlinedocs/gcc-9.2.0/gcc/Function-Attributes.html

void __attribute__((sysv_abi)) foo_system_v_abi(char a, StructType b, int c, StructType d, long e)
{
	/*
	 * Prema System V ABI konvenciji u skladu sa kojom (na osnovu atributa funkcije)
	 * vrsi prevodjenje posmatrane funkcije vazi sledece:
	 *
	 * Prvi parametar (char a) moze da se prosledi kroz registar
	 *     ---> prosledjuje se kroz prvi naredni slobodni registar (rdi) i zauzima citav registar
	 *
	 * Drugi parametar (StructType b) ne moze da se prosledi kroz registar
	 *     ---> prosledjuje se preko steka
	 *
	 * Treci parametar (int c) moze da se prosledi kroz registar
	 *     ---> prosledjuje se kroz prvi naredni slobodni registar (rsi) i zauzima citav registar
	 *
	 * Cetvrti parametar (StructType d) ne moze da se prosledi kroz registar
	 *     ---> prosledjuje se preko steka
	 *
	 * Peti parametar (long e) moze da se prosledi kroz registar
	 *     ---> prosledjuje se kroz prvi naredni slobodni registar (rdx)
	 */
}

void __attribute__((ms_abi)) foo_microsoft_abi(char a, StructType b, int c, StructType d, long e)
{
	/*
	 * Prema Microsoft ABI konvenciji u skladu sa kojom (na osnovu atributa funkcije)
	 * vrsi prevodjenje posmatrane funkcije vazi sledece:
	 *
	 * Prvi parametar (char a) moze da se prosledi kroz registar
	 *     ---> prosledjuje se kroz prvi naredni slobodni registar (rcx) i zauzima citav registar
	 *
	 * Drugi parametar (StructType b) ne moze da se prosledi kroz registar
	 *     ---> zamenjuje se pokazivacem na prostor gde se nalazi argument
	 *          a dati pokazivac se prosledjuje kroz prvi naredni slobodni registar(rdx)
	 *
	 * Treci parametar (int c) moze da se prosledi kroz registar
	 *     ---> prosledjuje se kroz prvi naredni slobodni registar (r8) i zauzima citav registar
	 *
	 * Cetvrti parametar (StructType d) ne moze da se prosledi kroz registar
	 *     ---> zamenjuje se pokazivacem na prostor gde se nalazi argument
	 *          a dati pokazivac se prosledjuje kroz prvi naredni slobodni registar(r9)
	 *
	 * Peti parametar (long e) moze da se prosledi kroz registar
	 *     ---> prosledjuje se kroz prvi naredni slobodni registar, ali posto su
	 *          svi registri zauzeti mora da se prosledi preko steka
	 */
}

int main()
{
	printf("%ld\n", sizeof(StructType));

	StructType struct_object1 = { 'a', 'b', 3l, 4, 'e', 6, 'g' };
	StructType struct_object2 = { 'g', 'f', 4l, 3, 'c', 2, 'a' };

	foo_system_v_abi('a', struct_object1, 3, struct_object2, 5l);
	foo_microsoft_abi('a', struct_object1, 3, struct_object2, 5l);

	return 0;
}
